/* This is a simple test server program to process a stock quote        */
/* request from the stock quote client application (client.c).          */
/* Communications are done using HTTP / SOAP / XML.  This is an         */
/* extremely simple case is that the server exits after processing      */
/* a single quote request.  A real world server would continuously      */
/* process quote requests indefinitely from multiple clients..          */

#include <stdio.h>
#include <stdlib.h>
#include "rtxsrc/rtxSOAP.h"
#include "rtxsrc/rtxSocket.h"
#include "rtxsrc/rtxStreamMemory.h"
#include "rtxsrc/rtxStreamSocket.h"
#include "StockQuoteWSDL.h"

#define DEFAULT_URL "http://127.0.0.1:9123"

int main (int argc, char** argv)
{
   GetLastTradePrice_Input req;
   GetLastTradePrice_Output resp;

   OSCTXT	soapCtxt, decCtxt, encCtxt;
   OSSOAPCONN   soapConn;
   OSOCTET*     ptmpbuf = 0;
   OSIPADDR     ipAddr = OSIPADDR_LOCAL;
   OSRTSOCKET   listenSocket;
   const char*  url = DEFAULT_URL;
   int		i, stat;
   OSBOOL       verbose = FALSE, trace = TRUE;

   /* Process command line arguments */

   if (argc > 1) {
      for (i = 1; i < argc; i++) {
         if (!strcmp (argv[i], "-v")) verbose = TRUE;
         else if (!strcmp (argv[i], "-url")) {
            url = argv[++i];
         }
         else if (!strcmp (argv[i], "-notrace")) trace = FALSE;
         else {
            printf ("usage: server [-v] [-url <url>] [-notrace]\n");
            printf ("   -v    verbose mode: print trace info\n");
            printf ("   -url  bind URL http://host:port\n");
            printf ("   -notrace  do not display trace info\n");
            return 0;
         }
      }
   }

   /* Init context */

   stat = rtXmlInitContext (&soapCtxt);
   if (0 != stat) {
      printf ("Context initialization failed.\n");
      rtxErrPrint (&soapCtxt);
      return stat;
   }

   if (verbose) 
      rtxSetDiag (&soapCtxt, 1);

   rtxSoapInitConn (&soapConn, &soapCtxt, OSSOAPV1, url);

   /* One-time socket package initialization (required for Windows only) */

   stat = rtxSocketsInit();
   if (0 != stat) {
      printf ("Socket initialization failed.\n");
      rtxErrSetData (&soapCtxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&soapCtxt);
      return stat;
   }

   /* Create listener socket */

   if ((stat = rtxSocketCreate (&listenSocket)) != 0) {
      rtxErrSetData (&soapCtxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&soapCtxt);
      return stat;
   }

   /* Bind the socket */

   if ((stat = rtxSocketStrToAddr (soapConn.host, &ipAddr)) != 0) {
      rtxErrSetData (&soapCtxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&soapCtxt);
      return stat;
   }
   if ((stat = rtxSocketBind (listenSocket, ipAddr, soapConn.port)) != 0) {
      rtxErrSetData (&soapCtxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&soapCtxt);
      return stat;
   }

   /* Wait for connection */

   if ((stat = rtxSocketListen (listenSocket, 1)) != 0) {
      rtxErrSetData (&soapCtxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&soapCtxt);
      return stat;
   }

   if (trace) printf ("Wait for connection...\n");

   stat = rtxSoapAcceptConn (listenSocket, &soapConn);
   if (0 != stat) {
      rtxErrSetData (&soapCtxt, stat, __FILE__, __LINE__);
      rtxErrPrint (&soapCtxt);
      return stat;
   }

   /* Wait for request */

   stat = rtxSoapRecvHttp (&soapConn);
   if (stat != 0) {
      rtxErrPrint (&soapCtxt);
      return stat;
   }
   if (trace) printf ("HTTP request received\n");

   /* Read HTTP request into memory.  This is the XML message to be     */
   /* decoded..                                                         */

   stat = rtxSoapRecvHttpContent (&soapConn, &ptmpbuf);
   if (stat != 0) {
      rtxErrPrint (&soapCtxt);
      return stat;
   }

   /* Since we know this server only processes stock quote requests and */
   /* nothing else, we can attempt to directly decode the request.      */
   /* If the server could process other message types, we would         */
   /* need to first determine what we got (by using the soap action     */
   /* field for example) before we could proceed..                      */

   stat = rtXmlInitContext (&decCtxt);
   if (0 != stat) {
      printf ("Context initialization failed.\n");
      rtxErrPrint (&decCtxt);
      return stat;
   }

   if (verbose) 
      rtxSetDiag (&decCtxt, 1);

   stat = rtxStreamMemoryAttach 
      (&decCtxt, ptmpbuf, soapConn.contentLength, OSRTSTRMF_INPUT);

   if (0 != stat) {
      printf ("rtxStreamMemoryAttach failed\n");
      rtxErrPrint (&decCtxt);
      return stat;
   }

   stat = XmlD_GetLastTradePrice_Input (&decCtxt, &req);

   if (0 == stat) {
      if (trace) {
         printf ("Decode was successful\n");
         printf ("Decoded record:\n");
         Print_GetLastTradePrice_Input 
            ("GetLastTradePriceInput", &req);
      }
   }
   else if (trace) {
      printf ("decode failed\n");
      rtxErrPrint (&decCtxt);
      return stat;
   }

   rtxStreamClose (&decCtxt);
   rtxFreeContext (&decCtxt);

   /* Init encode context */

   stat = rtXmlInitContext (&encCtxt);
   if (0 != stat) {
      printf ("Context initialization failed.\n");
      rtxErrPrint (&encCtxt);
      return stat;
   }

   if (verbose) rtxSetDiag (&encCtxt, 1);

   /* Prepare a response to the request */

   Init_GetLastTradePrice_Output (&encCtxt, &resp);
   resp.TradePrice.price = 34.50;

   /* Encode */

   stat = rtXmlSetEncBufPtr (&encCtxt, 0, 0);

   if (0 == stat)
      stat = XmlE_GetLastTradePrice_Output (&encCtxt, &resp);

   if (0 != stat) {
      printf ("Encoding failed\n");
      rtxErrPrint (&encCtxt);
      return stat;
   }

   /* Send response */
   soapConn.responseStatusCode = 200;
   stat = rtxSoapSendHttpResponse (&soapConn, rtXmlGetEncBufPtr(&encCtxt));
   if (stat != 0) {
      rtxErrPrint (&soapCtxt);
      return stat;
   }
   if (trace) printf ("HTTP response sent\n");

   /* Close connection */

   rtxStreamClose (&soapCtxt);
   rtxFreeContext (&soapCtxt);
   rtxFreeContext (&encCtxt);

   return stat;
}
