/*
 * Copyright (c) 2017 Objective Systems, Inc.
 *
 * See the COPYING.txt file in this subdirectory for conditions on use
 * of this software.  This notice applies to software only in this
 * subdirectory (oortsrc) and not to other software that may be a
 * a part of this distribution.
 *
 *****************************************************************************/
/**
 * @file ooLinkedList.h
 * Implementation of a doubly-linked list.
 */
#ifndef _OOLINKEDLIST_H_
#define _OOLINKEDLIST_H_

#include "oortsrc/ooCommon.h"

typedef struct _OOLinkedListNode {
   const void* data;
   struct _OOLinkedListNode* next;
   struct _OOLinkedListNode* prev;
} OOLinkedListNode;

typedef struct _OOLinkedList {
   OOUINT32 count;
   OOLinkedListNode* head;
   OOLinkedListNode* tail;
} OOLinkedList;

#ifdef __cplusplus
extern "C" {
#endif
/**
 * @defgroup llfuns Linked List Utility Functions
 *
 * Linked list implementation.
 * @{
 */
/**
 * This function appends an item to the linked list structure. The data item is
 * passed into the function as a void pointer that can point to any object of
 * any type. The malloc function is used to allocate memory for the
 * list node structure; therefore, all internal list memory will be released
 * whenever ooLinkedListFreeNodes is called. The pointer to the data item
 * itself is stored in the node structure - a copy is not made.
 *
 * @param pList        A pointer to a linked list structure onto which the data
 *                     item is to be appended. A pointer to an updated linked
 *                     list structure.
 * @param pData        A pointer to a data item to be appended to the list.
 * @return             A pointer to an allocated node structure used to link
 *                     the given data value into the list.
 */
EXTERN OOLinkedListNode*
ooLinkedListAppend (OOLinkedList* pList, void* pData);

/**
 * This function initializes a doubly linked list structure. It sets the
 * number of elements to zero and sets all internal pointer values to NULL.
 * A doubly linked-list structure is described by the OOLinkedList type.
 * Nodes of the list are of type OOLinkedListNode.
 *
 * @param pList  Pointer to a linked list structure to be initialized.
 */
EXTERN void ooLinkedListInit (OOLinkedList* pList);

/**
 * This function removes all nodes from the linked list and releases
 * the memory that was allocated for storing the node structures
 * (OOLinkedListNode). The data will not be released.
 *
 * @param pList  Pointer to linked list structure.
 */
EXTERN void ooLinkedListFreeNodes (OOLinkedList* pList);

/**
 * This function removes all nodes from the linked list structure and
 * releases the memory that was allocated for storing the node structures
 * (OOLinkedListNode) and for data.
 *
 * @param pList  Pointer to a linked list structure.
 */
EXTERN void ooLinkedListFreeAll (OOLinkedList* pList);

/**
 * This function inserts an item into the linked list structure before the
 * specified element.
 *
 * @param pList		A pointer to a linked list structure into which the
 *                        data item is to be inserted.
 * @param node          The position in the list where the item is to be
 *                        inserted.  The item will be inserted before this
 *                        node or appended to the list if node is null.
 * @param pData		A pointer to the data item to be inserted to the list.
 * @return		A pointer to an allocated node structure used to
 *                        link the given data value into the list.
 */
EXTERN OOLinkedListNode* ooLinkedListInsertBefore
(OOLinkedList* pList, OOLinkedListNode* node, const void* pData);

/**
 * This function inserts an item into the linked list structure after the
 * specified element.
 *
 * @param pList		A pointer to a linked list structure into which the
 *                        data item is to be inserted.
 * @param node          The position in the list where the item is to be
 *                        inserted.  The item will be inserted after this
 *                        node or added as the head element if node is null.
 * @param pData		A pointer to the data item to be inserted to the list.
 * @return		A pointer to an allocated node structure used to
 *                        link the given data value into the list.
 */
EXTERN OOLinkedListNode* ooLinkedListInsertAfter
(OOLinkedList* pList, OOLinkedListNode* node, const void* pData);

/**
 * This function removes a node containing the given data item from the
 * linked list structure.  The pointer to the removed data item is returned.
 *
 * @param pList  Pointer to linked list structure.
 * @param pdata  Pointer to the data item that is to be removed.
 */
EXTERN void* ooLinkedListRemove (OOLinkedList* pList, void* pdata);

/**
 * This function removes a node from the linked list structure.  A
 * pointer to the removed data item is returned.
 *
 * @param pList  Pointer to linked list structure.
 * @param node   Pointer to the node that is to be removed.
 */
EXTERN void* ooLinkedListRemoveNode
(OOLinkedList* pList, OOLinkedListNode* node);

/**
 * @}
 */
#ifdef __cplusplus
}
#endif
#endif
