#!/usr/bin/perl
# Execute all sample programs in given subdirectory

use File::Spec;
use Cwd;

require "./getMakeProg.pl";

select STDERR; $| = 1;
select STDOUT; $| = 1;

# include the current directory so that we can run reader and writer programs
# without modification
my $ps = (index ($ENV{"PATH"}, ';') >= 0) ? ';' : ':';
$ENV{"PATH"} = $ENV{"PATH"} . $ps . '.';
$ENV{"LD_LIBRARY_PATH"} = $ENV{"LD_LIBRARY_PATH"} . $ps . "../../lib";
if (defined ($ENV{"DYLD_LIBRARY_PATH"})) {
    $ENV{"DYLD_LIBRARY_PATH"} = $ENV{"DYLD_LIBRARY_PATH"} . $ps . "../../lib";
}
else {
    $ENV{"DYLD_LIBRARY_PATH"} = "../../lib";
}
my $WINDOWS = defined $ENV{"WINDIR"};

sub usage {
    $subdirline = shift;
    print "usage: sample.pl <subdir> <function> [<options>]\n";
    print "$subdirline\n";
    print "  <function>     function (test or clean)\n";
    print "  [<options>]:\n";
    print "    [-json]        run JSON sample programs\n";
    print "    [-make <make>] name of make utility to invoke (ex. nmake)\n";
    print "    [-nolic]       do not test against license-checked libraries\n";
    print "    [-nostop]      do not quit on error\n";
    print "    [-v]           enable verbose debug output\n";
    exit(1);
}

# Figure out what architecture this kit is based on:
# lnx - not Windows
# w32 - has 32-bit Windows libraries
# x64 - has 64-bit Windows libraries
# win - has both 32-bit and 64-bit Windows libraries
$architecture = "w32";
$architecture = "x64" if ((-e "c_64") && (!-e "c"));
$architecture = "win" if ((-e "c_64") && (-e "c"));
$architecture = "lnx" if (!$WINDOWS);
if ($architecture eq "w32") {
   $subdirline = "  <subdir>       subdirectory (c, cpp, csharp, java)";
}
elsif ($architecture eq "x64") {
   $subdirline = "  <subdir>       subdirectory (c_64, cpp_64, csharp, java)";
}
elsif ($architecture eq "win") {
   $subdirline = "  <subdir>       subdirectory (c, c_64, cpp, cpp_64, csharp, java)";
}
elsif ($architecture eq "lnx") {
   $subdirline = "  <subdir>       subdirectory (c, cpp, java)";
}
if ($#ARGV < 1) {
   usage($subdirline);
}

my $subdir = shift @ARGV;
my $function = shift @ARGV;
my $json = 0;
my $nostop = 0;
my $nerror = 0;
my $verbose = 0;
my $makeprog;
my $sampleDir = "sample";
my $liccheck = $function eq 'test' ? 1 : 0;

while ($_ = shift @ARGV) {
   if ($_ eq '-make') {
      usagew32() unless $makeprog = shift @ARGV;
      next;
   } 
   elsif ($_ eq '-nostop') {
      $nostop = 1;
   } 
   elsif ($_ eq '-json') {
      $json = 1;
      $sampleDir = "sample_json";
   } 
   elsif ($_ eq '-nolic') {
       $liccheck = 0;
   }
   elsif ($_ eq '-v') {
      $verbose = 1;
   } 
   else {
      print STDERR "Unknown option '$_'\n";
      usagew32();
   }
}

print ("Windows?: ", $WINDOWS ? "true" : "false", "\n") if ( $verbose );

if ($architecture eq "w32") {
   die ("$subdir is not a known subdirectory (must be c, cpp, csharp, or java)")
       unless ($subdir eq "c" || $subdir eq "cpp" || $subdir eq "csharp" || 
               $subdir eq "java");
}
elsif ($architecture eq "x64") {
   die ("$subdir is not a known subdirectory (must be c_64, cpp_64, csharp, or java)")
       unless ($subdir eq "c_64" || $subdir eq "cpp_64" || $subdir eq "csharp" || 
               $subdir eq "java");
}
elsif ($architecture eq "win") {
   die ("$subdir is not a known subdirectory (must be c, c_64, cpp, cpp_64, csharp, or java)")
       unless ($subdir eq "c" || $subdir eq "c_64" || $subdir eq "cpp" || $subdir eq "cpp_64" ||
               $subdir eq "csharp" || $subdir eq "java");
}
elsif ($architecture eq "lnx") {
   die ("$subdir is not a known subdirectory (must be c, cpp, or java)")
       unless ($subdir eq "c" || $subdir eq "cpp" || $subdir eq "java");
}

die ("$function is not a known function (must be test or clean)") 
    unless ($function eq "test" || $function eq "clean");

if ($json) {
    if ($architecture ne "win") {
       if (($architecture eq "w32") || ($architecture eq "lnx")) {
           $cdir = "c";
       }
       else {
           $cdir = "c_64";
       }
       die ("$subdir for JSON samples must be $cdir") unless ($subdir eq $cdir);
    }
    else {
        unless (($subdir eq "c") || ($subdir eq "c_64")) {
            die ("$subdir for JSON samples must be c or c_64");
        }
    }
}

$makeprog = getMakeProg() unless $makeprog;

chdir ("./$subdir") or die ("could not chdir to ./$subdir");

# handle dynamic link libraries on mac os X
$bmac = 0;
$bredirectok = 1;
if ($^O =~ /darwin/) {
   $bmac = 1;
   # macOS 11 (Big Sur) seems to have a lot of problems with redirection.
   $pv = `sw_vers -productVersion`;
   chomp($pv);
   @pieces = split(/\./, $pv);
   $major = $pieces[0];
   $bredirectok = 0 if ($major >= 11);
   $tmpdir = `pwd`;
   chomp($tmpdir);
   $ENV{'DYLD_FALLBACK_LIBRARY_PATH'} = "$tmpdir/lib"; 
}

# rename library directories in order to run sample against license checked
# libraries (may be overridden using -nolic option)
$swappedLibs = 0;
if ( $liccheck ) {
   if ($subdir =~ /^c/ && $subdir ne 'csharp') {
      if (-e 'lib' && -e 'lib_lic') {
         `mv lib lib_nolic`;
         `mv lib_lic lib`;
         print 'moved lib -> lib_nolic, lib_lic -> lib' . "\n";
         $swappedLibs = 1;
      }
   }
   elsif ( $subdir eq 'java') {
      if ( -e 'xbrt.jar' && -e 'xbrt_eval.jar' ) {
         `mv xbrt.jar xbrt_nolic.jar`;
         `mv xbrt_eval.jar xbrt.jar`;
         print "moved xbrt.jar -> xbrt_nolic.jar, xbrt_eval.jar -> xbrt.jar\n";
         $swappedLibs = 1;
      }
   }
   elsif ( $subdir eq 'csharp') {
      if ( -e 'xbrt.dll' && -e 'xbrt_eval.dll' ) {
         `mv xbrt.dll xbrt_nolic.dll`;
         `mv xbrt_eval.dll xbrt.dll`;
         print "moved xbrt.dll -> xbrt_nolic.dll, xbrt_eval.dll -> xbrt.dll\n";
         $swappedLibs = 1;
      }      
   }
}

execSamples ($sampleDir, $function);

# restore library directories
if ($swappedLibs) {
   if ($subdir =~ /^c/ && $subdir ne 'csharp') {
       `mv lib lib_lic`;
       `mv lib_nolic lib`;
       print "lib directories restored to original state\n";
   }
   elsif ( $subdir eq 'java') {
      `mv xbrt.jar xbrt_eval.jar`;
      `mv xbrt_nolic.jar xbrt.jar`;
      print "jar files restored to original state\n";
   }
   elsif ( $subdir eq 'csharp') {
      `mv xbrt.dll xbrt_eval.dll`;
      `mv xbrt_nolic.dll xbrt.dll`;
      print "C# dll files restored to original state\n";
   }
}

exit $nerror;

sub execSamples {
   my ($dirName) = @_;

   return 1 unless -e $dirName;

   my $build = "build";
   my $writer = "writer";
   my $reader = "reader";

   if ( $WINDOWS ) {
      $build = $build . ".bat";
      if ( $subdir eq "java" ) {
         $writer = $writer . ".bat";
         $reader = $reader . ".bat";
      }
      else {
         $writer = $writer . ".exe";
         $reader = $reader . ".exe";
      }
   }
   elsif ( $subdir eq "java" ) {
      $build = $build . ".sh";
      $writer = $writer . ".sh";
      $reader = $reader . ".sh";
   }

   # else: no extension for non-Windows, non-Java

    chdir ($dirName);
    @dirlist = (`ls .`);

    foreach $filename (@dirlist) {
        $filename =~ s/\n$//;

        if (-d $filename) {
            chdir ($filename);
   
            if ( $function eq "test" ) {
               my $runmake = -e "makefile";
               my $runbuild = !$runmake && -e $build;

               if ($runmake || $runbuild) {
                   printTestName($filename);
                   # Is this directory in the skip list?
                   if (checkSkipList()) {
                       print "SKIPPED";
                   }
                   else {
                      if ($runmake) {
                         execCmd($makeprog) or goto ERR;
                      }
                      else {
                         execCmd($build) or goto ERR;
                      }
                      my $ntest = 0;

                      if (-e $writer) {
                          print ("execute writer..\n") if ($verbose);
                          if ($filename =~ /Dll$/) {
                              execCmd ("runexe.bat writer > writer.log")
                                  or goto ERR;
                          }
                          else {
                              if ($bredirectok) {
                                 execCmd ("$writer > writer.log") or goto ERR;
                              }
                              else {
                                 execCmd ("$writer") or goto ERR;
                              }
                          }
                          $ntest++;
                      }

                       if (-e $reader) {
                          print ("execute reader..\n") if ($verbose);
                          if ($filename =~ /Dll$/) {
                              execCmd ("runexe.bat reader > reader.log")
                                  or goto ERR;
                          }
                          else {
                              if ($bredirectok) {
                                 execCmd ("$reader > reader.log") or goto ERR;
                              }
                              else {
                                 execCmd ("$reader") or goto ERR;
                              }
                          }
                          $ntest++;
                       }

                       if ((-e "server.exe" || -e "server") && 
                           (-e "client.exe" || -e "client")) {
                           $ntest++;
                           execClientServer() or goto ERR;
                       }

                       print "!" unless $ntest;
                    }
                   ERR: print "\n";
               }
            }
            elsif ($function eq "clean") {
                if (-e "makefile") {
                    execCmd ("$makeprog clean");
                }
                elsif ($subdir eq "java") {
                    `rm -rf bin`;
                    `rm -rf gensrc`;
                }
            }
            chdir ('..');

            last if (!$nostop && $nerror > 0);
        }
    }

    chdir ('..');
}

BEGIN {

    my $loadSkipList = sub {
        my $sampledir = getcwd();
        my $skipfile = "$sampledir/skiplist.inc";
        my @list;
        if (-e $skipfile) {
            open SKIP, $skipfile or die "Unable to open '$skipfile': $!";
            foreach (<SKIP>) {
                chomp;
                next if /^\s*$/;
                next if /^#/;
                    @list[++$#list] = $_;
            }
            close SKIP;
        }
        return @list;
    };
    my @skip = &$loadSkipList();

    sub checkSkipList {
        # Is this directory in the skip list?
        $cwd = getcwd();
        foreach $dir (@skip) {
            my $l = length($dir);
            next if $l == 0; 
            if (length($cwd)>$l) {
                if (substr($cwd, -$l, $l) eq $dir) {
                    return 1;
                }
            }
        }
        return 0;
    }
}

sub printTestName {
    local ($name) = shift @_;
    printf("$name%" . (30 - length($name)) . "s", '');
}

sub error {
   my $cmd = shift @_;
   
   $nerror++;

   if ($nostop) {
      print " ERROR cmd=$cmd";
   } 
   else {
      my $cwd;
      chomp($cwd = `pwd`);
      print "\n\n### TEST FAILURE ###\n";
      print "Command: $cmd\n";
      print "Directory: $cwd\n\n";
      # exit 1;
   }
}

sub execCmd {
    local ($cmd) = @_;
    my $devnull = File::Spec->devnull();

    $cmd =~ tr|/|\\| if ($ENV{'OS'} =~ /Windows/i && 
                         !$ENV{'TERM'} =~ /cygwin/);

    if ($verbose) {
        print "\n$cmd \n";
        print `$cmd`;
    }
    else {
        print (".");
        if ($bredirectok) {
           if (index($cmd, '>') >= 0) {
               system($cmd);
           }
           else {
               my $_cmd = "$cmd > $devnull";
               $_cmd .= " 2> $devnull" if  $nostop;
               system($_cmd);
           }
        }
        else {
           `$cmd`;
        }
    }

    $rc = $?;
    if ($rc != 0) {
        if($verbose) { printf "FAIL\n"; }
        error($cmd);
        return 0;
    }
    if($verbose) { printf "OK\n"; }
    return 1;
}

sub execDiff {
    local ($file1, $file2) = @_;
    if (-e $file1 && -e $file2) {
        my $cmd = "diff -w $file1 $file2";
        my $out;
        
        if ($verbose) {
            print "\nCommand: $cmd\n";
        } 
        else {
            print (".");
        }
        $out = `$cmd`;
        if (length($out) > 0) {
            error($cmd);
            return 0;
        }
    }
    return 1;
}

sub execGoodDiff {
    local $file = shift;
    local $goodFile = './good/' . $file;
    return &execDiff ($file, $goodFile);
}

sub execClientServer {
  
   my $child_pid;
   my $server_rc;
   my $client_rc;
   my $server_cmd = "server";
   my $client_cmd = "client";

   my $cmd = "$server_cmd & $client_cmd";
   if ($verbose) {
       print "\nCommand: $cmd\n";
   } 
   else {
       print (".");
   }
  
   if (!defined($child_pid = fork())) {
      die "cannot fork: $!";
   
   }
   elsif ($child_pid) {
       if ($verbose) {
           print "starting server..\n";
           print `$server_cmd`;
       }
       else { `$server_cmd`; }
       $server_rc = $?;

       # wait for child to return
       waitpid($child_pid, 0);
       $client_rc = $?;
   } 
   else {
       sleep 1; # give some time to the server

       if ($verbose) {
           print "starting client..\n";
           print `$client_cmd`;
       }
       else { `$client_cmd`; }

       # exit with appropriate return code
       exit ($? != 0);
   } 
	
   if (($server_rc != 0) || ($client_rc != 0)) {
       error($cmd);
       return 0;
   }
   return 1;

}
